@@ -1,7 +1,9 @@
 /* Copyright 2018, UCAR/Unidata See netcdf/COPYRIGHT file for copying
  * and redistribution conditions.*/
 /**
- * @file @internal The HDF4 file functions.
+ * @file
+ * @internal The HDF4 file functions. These provide a read-only
+ * interface to HDF4 SD files.
  *
  * @author Ed Hartnett
  */
@@ -67,6 +69,8 @@ hdf4_rec_grp_del(NC_GRP_INFO_T *grp)
          * scale. */
         if (hdf4_var->sdsid && SDendaccess(hdf4_var->sdsid) < 0)
             return NC_EHDFERR;
+
+        nullfree(hdf4_var);
     }
 
     return NC_NOERR;
@@ -239,7 +243,6 @@ nc4_set_var_type(nc_type xtype, int endianness, size_t type_size, char *type_nam
     type->endianness = endianness;
     type->size = type_size;
     type->hdr.id = (size_t)xtype;
-    type->hdr.hashkey = NC_hashmapkey(type->hdr.name, strlen(type->hdr.name));
 
     /* Return to caller. */
     *typep = type;
@@ -421,6 +424,8 @@ nc4_var_list_add_full(NC_GRP_INFO_T* grp, const char* name, int ndims, nc_type x
     if ((retval = nc4_set_var_type(xtype, endianness, type_size, type_name,
                                    &(*var)->type_info)))
         return retval;
+    /* Propate the endianness to the variable */
+    (*var)->endianness = (*var)->type_info->endianness;
 
     (*var)->type_info->rc++;
 
@@ -433,7 +438,10 @@ nc4_var_list_add_full(NC_GRP_INFO_T* grp, const char* name, int ndims, nc_type x
     }
 
     /* Var contiguous or chunked? */
-    (*var)->contiguous = contiguous;
+    if (contiguous)
+	(*var)->storage = NC_CONTIGUOUS;
+    else
+	(*var)->storage = NC_CHUNKED;
 
     /* Were chunksizes provided? */
     if (chunksizes)
@@ -577,45 +585,49 @@ hdf4_read_var(NC_FILE_INFO_T *h5, int v)
 }
 
 /**
- * @internal Open a HDF4 file.
+ * @internal Open a HDF4 SD file for read-only access.
  *
  * @param path The file name of the file.
  * @param mode The open mode flag.
  * @param basepe Ignored by this function.
  * @param chunksizehintp Ignored by this function.
  * @param parameters pointer to struct holding extra data (e.g. for
- * parallel I/O) layer. Ignored if NULL.
+ * parallel I/O) layer. Ignored if NULL. Ignored by this function.
  * @param dispatch Pointer to the dispatch table for this file.
- * @param nc_file Pointer to an instance of NC.
+ * @param nc_file Pointer to an instance of NC. The ncid has already
+ * been assigned, and is in nc_file->ext_ncid.
  *
  * @return ::NC_NOERR No error.
  * @return ::NC_EINVAL Invalid input.
+ * @return ::NC_EHDFERR Error from HDF4 layer.
+ * @return ::NC_ENOMEM Out of memory.
  * @author Ed Hartnett
  */
 int
 NC_HDF4_open(const char *path, int mode, int basepe, size_t *chunksizehintp,
-             void *parameters, NC_Dispatch *dispatch, NC *nc_file)
+             void *parameters, const NC_Dispatch *dispatch, int ncid)
 {
     NC_FILE_INFO_T *h5;
     NC_HDF4_FILE_INFO_T *hdf4_file;
+    NC *nc;
     int32 num_datasets, num_gatts;
     int32 sdid;
     int v, a;
-    NC_FILE_INFO_T* nc4_info = NULL;
-    int retval = NC_NOERR;
+    int retval;
 
     /* Check inputs. */
-    assert(nc_file && path);
+    assert(path);
 
     LOG((1, "%s: path %s mode %d params %x", __func__, path, mode, parameters));
 
+    /* Find pointer to NC. */
+    if ((retval = NC_check_id(ncid, &nc)))
+        return retval;
+
     /* Check the mode for validity */
     if (mode & ILLEGAL_OPEN_FLAGS)
         return NC_EINVAL;
 
-    /* Open the file. */
-    nc_file->int_ncid = nc_file->ext_ncid;
-
     /* Open the file and initialize SD interface. */
     if ((sdid = SDstart(path, DFACC_READ)) == FAIL)
         return NC_EHDFERR;
@@ -625,11 +637,9 @@ NC_HDF4_open(const char *path, int mode, int basepe, size_t *chunksizehintp,
         return NC_EHDFERR;
 
     /* Add necessary structs to hold netcdf-4 file data. */
-    if ((retval = nc4_nc4f_list_add(nc_file, path, mode)))
+    if ((retval = nc4_file_list_add(ncid, path, mode, (void **)&h5)))
         return retval;
-    nc4_info = NC4_DATA(nc_file);
-    assert(nc4_info && nc4_info->root_grp);
-    h5 = nc4_info;
+    assert(h5 && h5->root_grp);
     h5->no_write = NC_TRUE;
     h5->root_grp->atts_read = 1;
 
@@ -670,6 +680,7 @@ NC_HDF4_open(const char *path, int mode, int basepe, size_t *chunksizehintp,
  *
  * @return ::NC_NOERR No error.
  * @return ::NC_EBADID Bad ncid.
+ * @return ::NC_EHDFERR Error from HDF4 layer.
  * @author Ed Hartnett
  */
 int
@@ -686,6 +697,7 @@ NC_HDF4_abort(int ncid)
  *
  * @return ::NC_NOERR No error.
  * @return ::NC_EBADID Bad ncid.
+ * @return ::NC_EHDFERR Error from HDF4 layer.
  * @author Ed Hartnett
  */
 int
@@ -708,25 +720,15 @@ NC_HDF4_close(int ncid, void *ignore)
     if ((retval = hdf4_rec_grp_del(h5->root_grp)))
         return retval;
 
-    /* Delete all the list contents for vars, dims, and atts, in each
-     * group. */
-    if ((retval = nc4_rec_grp_del(h5->root_grp)))
-        return retval;
-
     /* Close hdf4 file and free HDF4 file info. */
     hdf4_file = (NC_HDF4_FILE_INFO_T *)h5->format_file_info;
     if (SDend(hdf4_file->sdid))
         return NC_EHDFERR;
     free(hdf4_file);
 
-    /* Misc. Cleanup */
-    nclistfree(h5->alldims);
-    nclistfree(h5->allgroups);
-    nclistfree(h5->alltypes);
-
-    /* Free the nc4_info struct; above code should have reclaimed
-       everything else */
-    free(h5);
+    /* Free the NC_FILE_INFO_T struct. */
+    if ((retval = nc4_nc4f_list_del(h5)))
+        return retval;
 
     return NC_NOERR;
 }
